from dataclasses import dataclass
from datetime import date
from pathlib import Path

from geopy.distance import geodesic  # type: ignore
from typing import TypeVar, ClassVar, FrozenSet, Any
from collections import defaultdict

from .constants import Auto, Volumes, Year
from . import events, commands
from steelo.utilities.global_variables import ENERGY_EMISSION_FACTORS, ACTIVE_STATUSES, SIMULATION_START_YEAR
import steelo.utilities.global_variables as gv
from ..utilities.utils import merge_two_dictionaries
import json

import logging

logger = logging.getLogger(__name__)


BoundedStringT = TypeVar("BoundedStringT", bound="BoundedString")


class BoundedString(str):
    _valid_values: ClassVar[FrozenSet[str]] = frozenset()  # Make it immutable

    def __new__(cls: type[BoundedStringT], value: str) -> BoundedStringT:
        if value not in cls._valid_values:
            raise ValueError(f"Invalid value for {cls.__name__}: {value}. Valid values are: {cls._valid_values}")
        return str.__new__(cls, value)

    @classmethod
    def is_valid(cls, value: str) -> bool:
        return value in cls._valid_values

    @classmethod
    def all_values(cls) -> FrozenSet[str]:
        return cls._valid_values


class ProductCategory(BoundedString):
    _valid_values = frozenset({"Flat", "Long", "Tubes", "Semi-finished"})


class TimeFrame:
    """
    Holds the start and end year for the model calculations.
    """

    def __init__(self, *, start: Year = Year(2025), end: Year = Year(2050)) -> None:
        self.start = start
        self.end = end

    @property
    def years(self) -> tuple[Year, ...]:
        return tuple(Year(y) for y in range(self.start, self.end + 1))


class PointInTime:
    """
    Holds the current year and the time frame for the model calculations.

    For example, the lifetime of a furnace group is a point in time. It
    started in a specific year and will end in a specific year and how
    much lifetime is left can be calculated based on the current year.
    """

    def __init__(self, *, current: Year = Auto, time_frame: TimeFrame = Auto) -> None:
        if not time_frame:
            time_frame = TimeFrame()
        if not current:
            current = time_frame.start
        self.current = current
        self.time_frame = time_frame

    @property
    def start(self) -> Year:
        return self.time_frame.start

    @property
    def end(self) -> Year:
        return self.time_frame.end

    @property
    def number_of_years(self) -> int:
        return self.end - self.start

    @property
    def remaining_number_of_years(self) -> int:
        return self.end - self.current

    @property
    def elapsed_number_of_years(self) -> int:
        return self.current - self.start

    @property
    def expired(self) -> bool:
        return self.current >= self.end


@dataclass
class Location:
    lat: float
    lon: float
    country: str
    region: str
    iso3: str | None = None
    distance_to_other_iso3: dict[str, float] | None = None


class PrimaryFeedstock:
    def __init__(self, metallic_charge: str, technology: str):
        self.name = f"{technology}_{metallic_charge}"
        self.metallic_charge = None
        self.technology = technology
        self.required_quantity_per_ton_of_product = None
        self.secondary_feedstock: dict[str, float] = {}
        self.energy_requirements: dict[str, float] = {}

    def __repr__(self) -> str:
        return self.name

    def add_secondary_feedstock(self, name: str, share: float) -> None:
        self.secondary_feedstock[name] = share

    def add_maximum_share_in_product(self, limit: float) -> None:
        self.maximum_share_in_product = limit

    def add_minimum_share_in_product(self, limit: float) -> None:
        self.minimum_share_in_product = limit

    def add_energy_requirement(self, vector_name: str, amount: float) -> None:
        self.energy_requirements[vector_name] = amount

    def add_share_constraint(self, constraint_type: str, value: float) -> None:
        assert constraint_type in ["Maximum", "Minimum", "Minium"], f"Invalid constraint type {constraint_type}"
        if constraint_type == "Maximum":
            self.add_maximum_share_in_product(value)
        else:
            self.add_minimum_share_in_product(value)


@dataclass
class Technology:
    name: str
    product: str
    technology_readiness_level: int | None = None
    process_emissions: float | None = None
    dynamic_business_case: list[PrimaryFeedstock] = Auto
    # TODO: all below are dynamic, based on feedstock:
    emissions_factor: float | None = None
    energy_consumption: float | None = None
    bill_of_materials: dict[str, dict[str, dict[str, Any]]] = Auto  # FIXME, dunno how this should look like
    lcop: float | None = None
    capex_type: str | None = None

    _allowed_names = {"EAF", "DRI", "DRI-EAF", "BF", "BOF", "BF-BOF", "OHF", "other", "unknown"}

    def __post_init__(self):
        # Check if the technology name is valid
        if self.name not in self._allowed_names:
            raise ValueError(f"Invalid name: {self.name}")
        if self.capex_type is None:
            self.capex_type = "greenfield"

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        # Technologies are equal if they have the same name
        return self.name == other.name

    def __repr__(self) -> str:
        return f"Technology: <{self.name}>"

    def emissions_per_unit(self) -> float:
        emissions_per_unit: float = 0.0
        for feedstock in self.dynamic_business_case:
            if feedstock.required_quantity_per_ton_of_product is None:
                continue

            for sec_fs in feedstock.secondary_feedstock:
                if sec_fs in ENERGY_EMISSION_FACTORS:
                    emissions_per_unit += (
                        feedstock.required_quantity_per_ton_of_product
                        * feedstock.secondary_feedstock[sec_fs]
                        * ENERGY_EMISSION_FACTORS[sec_fs]
                    )
            for en_req in feedstock.energy_requirements:
                # Todo: this is probably wrong though:
                if en_req == "Flexible":
                    emissions_per_unit += (
                        feedstock.required_quantity_per_ton_of_product
                        * feedstock.energy_requirements[en_req]
                        * ENERGY_EMISSION_FACTORS["Natural gas"]
                    )
                if en_req in ENERGY_EMISSION_FACTORS:
                    emissions_per_unit += (
                        feedstock.required_quantity_per_ton_of_product
                        * feedstock.energy_requirements[en_req]
                        * ENERGY_EMISSION_FACTORS[en_req]
                    )
        return emissions_per_unit

    @property
    def technology_capex(self) -> float:
        from .Capex import FULL_GREENFIELD_CAPEX

        name_to_capex = {
            "greenfield": FULL_GREENFIELD_CAPEX,
            "brownfield": {k: v / 2 for k, v in FULL_GREENFIELD_CAPEX.items()},
        }
        if self.capex_type is None:
            raise ValueError("Capex type must not be None")
        capex_dict = name_to_capex.get(self.capex_type)
        if capex_dict is None:
            raise ValueError(f"Unknown capex_type: {self.capex_type}")
        return capex_dict.get(self.name, 0.0)

    def add_feedstock_to_dynamic_business_case(self, primary_feedstock: PrimaryFeedstock) -> None:
        self.dynamic_business_case.append(primary_feedstock)


@dataclass
class ProductionThreshold:
    """Production thresholds for the utilization rate of a furnace group."""

    low: float = 0.1
    high: float = 0.95


class FurnaceGroup:
    def __init__(
        self,
        *,
        furnace_group_id: str,
        capacity: Volumes,
        status: str,
        last_renovation_date: date,
        technology: Technology,
        historical_production: dict[Year, Volumes],
        utilization_rate: float,
        lifetime: PointInTime,
        production_threshold: ProductionThreshold = ProductionThreshold(),
        equity_share: float = 0.2,
        cost_of_debt: float = 0.05,
    ) -> None:
        self.furnace_group_id = furnace_group_id
        self.capacity = capacity
        self.status = status
        self.last_renovation_date = last_renovation_date
        self.technology = technology
        self.historical_production = historical_production
        self.utilization_rate = utilization_rate
        self.lifetime = lifetime
        self.production_threshold = production_threshold
        self.created_by_PAM = False

        # Economic variables
        self.equity_share = equity_share
        self.cost_of_debt = cost_of_debt

    def __repr__(self) -> str:
        return f"FurnaceGroup: <{self.furnace_group_id}>"

    def __eq__(self, other) -> bool:
        return self.furnace_group_id == other.furnace_group_id

    def __hash__(self):
        return hash(self.furnace_group_id)

    def set_technology(self, new_technology: Technology) -> None:
        self.technology = new_technology

    def update_renovation_date(self, new_renovation_date: date) -> None:
        self.last_renovation_date = new_renovation_date

    def renovation_needed(self, current_date: date, average_renovation_frequency: int = 20) -> bool:
        return (current_date.year - self.last_renovation_date.year) > average_renovation_frequency

    def update_utilization_rate(self, utilization_rate: Volumes) -> None:
        self.utilization_rate = utilization_rate

    def report_bill_of_materials(self):
        return {
            item_name: item_dict["demand"] * item_dict["unit_cost"]["Value"] * self.production
            for bom_group, items in self.technology.bill_of_materials.items()
            for item_name, item_dict in items.items()
        }

    @property
    def emissions(self):
        return self.production * self.technology.emissions_per_unit() / (1e6)  # Convert from gram to ton

    @property
    def fixed_opex(self) -> float:
        from .calculate_costs import calculate_fixed_opex

        return calculate_fixed_opex(self.technology.technology_capex)

    @property
    def opex(self) -> float:
        from .calculate_costs import calculate_opex, BillsOfMaterial

        return calculate_opex(
            bill_of_materials=getattr(BillsOfMaterial, self.technology.name.replace("-", "_")),
            fixed_opex=self.fixed_opex * self.capacity,
            production=self.production,
        )

    @property
    def total_investment(self) -> float:
        return self.capacity * self.technology.technology_capex

    @property
    def debt_repayment_per_year(self) -> list[float]:
        from .calculate_costs import calculate_debt_repayment

        """Returns the debt repayment amount for each year of the remaining lifetime."""
        return calculate_debt_repayment(
            total_investment=self.total_investment,
            equity_share=self.equity_share,
            lifetime=self.lifetime.number_of_years,
            cost_of_debt=self.cost_of_debt,
            lifetime_remaining=self.lifetime.remaining_number_of_years,
        )

    @property
    def debt_repayment_for_current_year(self) -> float:
        from .calculate_costs import calculate_current_debt_repayment

        """Returns the debt repayment amount for the current year."""
        if self.cost_of_debt is None:
            # FIXME how do we even end up here?
            self.cost_of_debt = 0.05
        return calculate_current_debt_repayment(
            total_investment=self.total_investment,
            lifetime_expired=self.lifetime.expired,
            lifetime_years=self.lifetime.number_of_years,
            years_elapsed=self.lifetime.elapsed_number_of_years,
            cost_of_debt=self.cost_of_debt,
            equity_share=self.equity_share,
        )

    @property
    def production(self) -> float:
        return self.utilization_rate * self.capacity

    @property
    def unit_production_cost(self) -> float:
        from .calculate_costs import calculate_unit_production_cost

        """Maybe divide by production to normalize to unit?"""
        return calculate_unit_production_cost(
            opex=self.opex, current_debt_repayment=self.debt_repayment_for_current_year, production=self.production
        )

    @property
    def is_underutilized(self) -> bool:
        return self.utilization_rate < self.production_threshold.low

    def optimal_technology_name(self, price) -> tuple[str, float]:  # TODO: @Marcus update implimentation
        from .calculate_costs import derive_best_technology_switch, stranding_asset_cost

        cosa = stranding_asset_cost(
            self.debt_repayment_per_year,
            [self.opex] * len(self.debt_repayment_per_year),
            self.lifetime.remaining_number_of_years,
            price,
            self.production,
            self.cost_of_debt,
        )

        # optimal_tech_name, _ = max(self.technology_switch_npv.items(), key=lambda item: item[1]
        npv, optimal_tech_name = derive_best_technology_switch(
            current_technology=self.technology.name,
            cost_of_debt=self.cost_of_debt,
            equity_share=self.equity_share,
            lifetime=gv.LIFETIME["steel"],
            expected_utilisation_rate=0.7,
            capacity=1000,
            price=price,
            mvp=True,
            cosa=cosa,
        )

        return optimal_tech_name, npv

    # def no_profitable_technology(self, price) -> bool:  # TODO @Marcus update this implimentatio
    #     """
    #     Assesses if there is no profitable technology to switch to
    #     """
    #     from .calculate_costs import derive_best_technology_switch

    #     derive_best_technology_switch(
    #         current_technology=self.technology.name,
    #         cost_of_debt=self.cost_of_debt,
    #         equity_share=self.equity_share,
    #         lifetime=gv.LIFETIME["steel"],
    #         expected_utilisation_rate=0.7,
    #         capacity=1000,
    #         price=price,
    #         mvp=True,
    #         cosa=cosa,
    #     )

    #     optimal_npv, optimal_tech = max(self.technology_switch_npv.items(), key=lambda item: item[1])
    #     threshold = 100
    #     return optimal_npv < threshold


class Plant:
    def __init__(
        self,
        *,
        plant_id: str,
        location: Location,
        furnace_groups: list[FurnaceGroup],
        power_source: str,
        soe_status: str,
        parent_gem_id: str,
        workforce_size: int,
        certified: bool,
        category_steel_product: set[ProductCategory],
    ) -> None:
        # ID:
        self.plant_id = plant_id

        # Location
        self.location = location

        # Plant operations
        self.furnace_groups = furnace_groups
        self.power_source = power_source

        # Company management info
        self.soe_status = soe_status  # State-owned plants expected to make different decisions than private companies (e.g., no country shift)
        self.parent_gem_id = parent_gem_id
        self.workforce_size = workforce_size  # Keep for OPEX calc and impacts

        # Product quality
        self.certified = certified  # Merge 'ISO 14001', 'ISO 50001', and 'ResponsibleSteel Certification' into one single col, True (one enough) / False
        self.category_steel_product = (
            category_steel_product  # Discuss with Hannah if demand is dependent on steel quality
        )

        # collect domain events
        self.events: list[events.Event] = []
        self.added_capacity = Volumes(0)
        self.removed_capacity = Volumes(0)

    def add_furnace_group(self, new_furnace_group: FurnaceGroup) -> None:
        self.furnace_groups.append(new_furnace_group)

    def get_consumption_values(self) -> None:
        """
        This will fetch the consumption values per production unit for the given tech stack
        We assume that the tech stack lives somewhere accessible to the plant agent.
        """

    def has_DRI_furnace(self) -> bool:
        """
        Check if the plant has a DRI furnace
        """
        for furnace_group in self.furnace_groups:
            if furnace_group.technology.name == "DRI":
                return True
        return False

    def set_technology_lcops(self, technology_lops: dict[str, float]) -> None:
        """
        sets levelised cost of production to the technologies in the plant
        """
        has_DRI = self.has_DRI_furnace()
        for furnace_group in self.furnace_groups:
            if furnace_group.technology.name == "EAF":
                if has_DRI:
                    furnace_group.technology.lcop = technology_lops["DRI-EAF"]
                else:
                    furnace_group.technology.lcop = technology_lops["EAF"]
            elif furnace_group.technology.name == "BOF":
                furnace_group.technology.lcop = technology_lops["Avg BF-BOF"]
            furnace_group.technology.lcop

    def calculate_average_steel_cost_and_capacity(self):
        """
        Calculates average steel cost and capacity for the plant for active steel furnaces
        """
        total_cost = 0
        total_capacity = 0
        for furnace_group in self.furnace_groups:
            if (
                furnace_group.status in ACTIVE_STATUSES
                and furnace_group.technology.product == "Steel"
                and furnace_group.technology.lcop is not None
            ):
                total_cost += furnace_group.technology.lcop * furnace_group.capacity
                total_capacity += furnace_group.capacity
        if total_capacity == 0:
            return
        self.average_steel_cost = total_cost / total_capacity
        self.steel_capacity = total_capacity

    def distance_to(self, location: Location) -> float:
        if self.location.distance_to_other_iso3 is not None and location.iso3 is not None:
            return self.location.distance_to_other_iso3[location.iso3]
        else:
            return geodesic((self.location.lat, self.location.lon), (location.lat, location.lon)).kilometers

    def __eq__(self, other) -> bool:
        return self.plant_id == other.plant_id

    def __hash__(self):
        return hash(self.plant_id)

    def __repr__(self) -> str:
        return f"Plant: <{self.plant_id}>"

    def get_furnace_group(self, furnace_group_id: str) -> FurnaceGroup:
        try:
            return next(fg for fg in self.furnace_groups if fg.furnace_group_id == furnace_group_id)
        except StopIteration:
            raise ValueError(f"Furnace group {furnace_group_id} not found in plant {self.plant_id}")

    def reset_capacity_changes(self):
        self.added_capacity = Volumes(0)
        self.removed_capacity = Volumes(0)

    def close_furnace_group(self, furnace_group_id: str) -> None:
        """
        Close a furnace group by mothballing it.

        Retrieves the furnace group with id==furnace_group_id, adds its capacity to the removed capacity,
        sets its capacity to zero, updates its status to "Mothballed", and logs a FurnaceGroupClosed event.
        """
        furnace_group = self.get_furnace_group(furnace_group_id)
        self.removed_capacity = Volumes(self.removed_capacity + furnace_group.capacity)
        furnace_group.capacity = Volumes(0)
        furnace_group.status = "Mothballed"
        self.events.append(events.FurnaceGroupClosed(furnace_group_id=furnace_group_id))

    def renovate_furnace_group(self, furnace_group_id: str) -> None:
        """
        Renovate a furnace group.

        Resets furnace group lifetime to a new 20-year period, changes technology CAPEX type to "brownfield", and sets last_renovation date.
        An event is logged to indicate that the furnace group has been renovated.
        """
        furnace_group = self.get_furnace_group(furnace_group_id)
        current_year = furnace_group.lifetime.current
        furnace_group.last_renovation_date = date(current_year, 1, 1)
        furnace_group.lifetime = PointInTime(
            current=current_year,
            time_frame=TimeFrame(start=current_year, end=Year(current_year + gv.LIFETIME["steel"])),
        )
        furnace_group.technology.capex_type = "brownfield"
        self.events.append(events.FurnaceGroupRenovated(furnace_group_id=furnace_group_id))

    def change_furnace_group_technology(self, furnace_group_id: str, technology_name: str) -> None:
        """
        Change the technology of a specified furnace group.

        Updates the technology of the furnace group with id==`furnace_group_id` to a new technology
        specified by `technology_name`. The method retrieves the corresponding bill of materials, updates
        the product and capex type, resets the furnace group's lifetime to a new 20-year period based on
        the current year, and appends a technology change event.

        TODO:
            - Ensure that the cost of debt is inherited based on location.
            Marcus? if we only change the technology, we should not change the cost of debt?
        """
        from .calculate_costs import BillsOfMaterial

        furnace_group = self.get_furnace_group(furnace_group_id)
        technology = Technology(
            name=technology_name,
            bill_of_materials=getattr(BillsOfMaterial, technology_name.replace("-", "_")),
            product=furnace_group.technology.product,
            capex_type="greenfield",
        )

        furnace_group.technology = technology
        current_year = furnace_group.lifetime.current
        furnace_group.lifetime = PointInTime(
            current=current_year,
            time_frame=TimeFrame(start=current_year, end=Year(current_year + gv.LIFETIME["steel"])),
        )
        self.events.append(events.FurnaceGroupTechChanged(furnace_group_id=furnace_group_id))

        # TODO: @Marcus - make sure that cost of debt is inherited based on location!

    def evaluate_furnace_group_strategy(self, furnace_group_id: str, price: float) -> commands.Command | None:
        """
        This function will evaluate the strategy for the given furnace group
        """
        furnace_group = self.get_furnace_group(furnace_group_id)
        if furnace_group.is_underutilized:  # or furnace_group.no_profitable_technology:
            return commands.CloseFurnaceGroup(plant_id=self.plant_id, furnace_group_id=furnace_group.furnace_group_id)

        optimal_tech_name, tech_npv = furnace_group.optimal_technology_name(price)
        if tech_npv < 0:
            logger.debug(
                f"Optimal tech is {optimal_tech_name} with npv of {tech_npv}, but also capacity of {furnace_group.capacity}"
            )
            return None
        technology_already_optimal = furnace_group.technology.name == optimal_tech_name

        # Check whether the asset should be renovated
        if furnace_group.lifetime.expired and technology_already_optimal:
            return commands.RenovateFurnaceGroup(
                plant_id=self.plant_id, furnace_group_id=furnace_group.furnace_group_id
            )
        # Check whether the asset should switch to a more profitable technology
        technology_not_optimal = not technology_already_optimal
        if technology_not_optimal:
            return commands.ChangeFurnaceGroupTechnology(
                plant_id=self.plant_id,
                furnace_group_id=furnace_group.furnace_group_id,
                technology_name=optimal_tech_name,
            )

        # If none of the above conditions are met, continue operations without changing strategy
        return None

    def generate_new_furnace(
        self,
        technology_name: str,
        product: str,
        current_year: int,
        cost_of_debt: float = 0.02,
        expected_util_rate: float = 0.07,
        capacity: int = 1000,
    ) -> FurnaceGroup:
        """
        Generate a new furnace group for the specified technology.

        This is a placeholder implementation that creates a FurnaceGroup object using the given
        technology name and current year. The technology's bill of materials is retrieved from
        the BillsOfMaterial container. The furnace group is initialized with default values for
        capacity, status, utilization, and lifetime.

        Parameters:
            technology_name (str): The name of the technology.
            current_year (int): The current year used to set the furnace group's timeline.

        Returns:
            FurnaceGroup: The newly created furnace group.
        """
        from .calculate_costs import BillsOfMaterial

        technology = Technology(
            name=technology_name,
            bill_of_materials=getattr(BillsOfMaterial, technology_name.replace("-", "_")),
            product=product,
        )

        new_furnace_id = self.get_new_furnance_id_number()
        active_year = current_year + 3  # TODO: update with years to add given technology being built

        furnace_group = FurnaceGroup(
            furnace_group_id=new_furnace_id,
            capacity=Volumes(capacity),
            status="announced",
            last_renovation_date=date(current_year, 1, 1),
            technology=technology,
            historical_production={},
            utilization_rate=expected_util_rate,
            cost_of_debt=cost_of_debt,
            lifetime=PointInTime(
                current=Year(current_year),
                time_frame=TimeFrame(start=Year(active_year), end=Year(active_year + gv.LIFETIME["steel"])),
            ),
        )
        furnace_group.created_by_PAM = True

        return furnace_group

    def evaluate_expansion(self, environment, new_furnace: FurnaceGroup) -> commands.Command | None:
        """
        Evaluate whether a new furnace group should be added to the plant.

        Compares the predicted market price with 150% of the new furnace group's unit production cost.
        If the predicted price exceeds this threshold, returns an AddFurnaceGroup command; otherwise, returns None.

        Parameters:
            environment: An object that provides market information and demand, including a method
                        to predict new market prices.
            new_furnace (FurnaceGroup): The furnace group under evaluation for expansion.

        Returns:
            commands.Command | None: The command to add the furnace group if expansion is justified, otherwise None.
        """
        predicted_price = environment._predict_new_market_price(
            demand=environment.current_demand,
            new_furnace_group=new_furnace,
        )

        if predicted_price > new_furnace.unit_production_cost * 1.5:  # 50% margin TODO: improve expansion logic
            return commands.AddFurnaceGroup(
                plant_id=self.plant_id, furnace_group_id=new_furnace.furnace_group_id, new_furnace=new_furnace
            )
        return None

    def furnace_group_added(self, furnace_group_id: str, plant_id: str) -> None:
        self.events.append(events.FurnaceGroupAdded(furnace_group_id=furnace_group_id, plant_id=plant_id))

    def report_cost_breakdown(self):
        """
        Generate a cost breakdown report for each technology.

        Iterates over furnace groups, calculates cost components (debt, interest, O&M, production),
        and merges the bill of materials for each technology. Returns a dictionary mapping technology
        names to their cost breakdown details.

        Returns:
            dict: A dictionary with cost breakdown for each technology.
        """
        from .calculate_costs import calculate_debt_report

        cost_breakdown_by_tech = defaultdict(lambda: defaultdict(float))
        bom_breakdown_by_tech = defaultdict(lambda: defaultdict(float))
        final_breakdown = {}

        # Iterate over each furnace group to update the cost breakdown
        for fg in self.furnace_groups:
            if fg.capacity == 0:
                continue
            tech = fg.technology.name
            # Calculate cost components
            principal_debt, interest = calculate_debt_report(
                total_investment=fg.total_investment,
                lifetime_expired=fg.lifetime.expired,
                equity_share=fg.equity_share,
                lifetime_years=fg.lifetime.number_of_years,
                years_elapsed=fg.lifetime.elapsed_number_of_years,
                cost_of_debt=fg.cost_of_debt,
            )
            cost_breakdown_by_tech[tech]["Principal_debt"] += principal_debt
            cost_breakdown_by_tech[tech]["Interest"] += interest
            cost_breakdown_by_tech[tech]["O&M"] += fg.fixed_opex * fg.production
            cost_breakdown_by_tech[tech]["Production"] += fg.production

            # Merge Bill of Materials for this technology
            bill_of_materials = fg.report_bill_of_materials()  # This should be a dict
            if bill_of_materials:
                merge_two_dictionaries(bom_breakdown_by_tech[tech], bill_of_materials)

            # Integrate each technology's merged Bill of Materials into its cost breakdown
            for tech, cost_data in cost_breakdown_by_tech.items():
                cost_data["Bill of Materials"] = dict(bom_breakdown_by_tech[tech])
                final_breakdown[tech] = dict(cost_data)

        return final_breakdown

    def aggregate_emissions(self):
        """
        Report the total emissions for the plant.
        """
        emissions = 0
        for fg in self.furnace_groups:
            emissions += fg.emissions
        return emissions

    def get_new_furnance_id_number(self) -> str:
        """
        Get a furnace group id for a potential new furnace group.
        Adds 1 to the number of the last furnace group id belonging to the plant.
        """
        try:
            newest_furnace = self.furnace_groups[-1].furnace_group_id
        except IndexError:
            # No furnace groups exist yet, so we create a new one
            return f"{self.plant_id}_1"

        # This is completely nuts - FIXME please!
        # Handle different ID formats (fg_group_N or plant_id_N)
        parts = newest_furnace.split("_")
        if len(parts) >= 2 and parts[0] == "fg":
            # Format: fg_group_N
            try:
                if parts[1] == "group":
                    new_number = int(parts[2]) + 1
                    return f"fg_group_{new_number}"
                else:
                    # Try to get the number from the last part
                    new_number = int(parts[-1]) + 1
                    return f"{parts[0]}_{parts[1]}_{new_number}"
            except (IndexError, ValueError):
                # Fallback if we can't parse the ID format
                return f"fg_group_{len(self.furnace_groups) + 1}"
        else:
            # Format: plant_id_N
            try:
                new_number = int(parts[-1]) + 1
                return f"{self.plant_id}_{new_number}"
            except (IndexError, ValueError):
                # Fallback if we can't parse the ID format
                return f"{self.plant_id}_{len(self.furnace_groups) + 1}"


class PlantGroup:
    def __init__(self, *, plant_group_id: str, plants: list[Plant]) -> None:
        self.plant_group_id = plant_group_id
        self.plants = plants

    # TODO: What else goes in here?
    # Maybe this is in reality where the consideration of adding a furnace comes into play.
    # According to Artem it will be the parent group that assess the need an opportunity for a new furnace.


class DemandCenter:
    def __init__(self, demand_center_id: str, center_of_gravity: Location, demand_by_year: dict[Year, Volumes]) -> None:
        self.demand_center_id = demand_center_id
        self.center_of_gravity = center_of_gravity
        self.demand_by_year = demand_by_year

    def __repr__(self) -> str:
        return f"DemandCenter: <{self.demand_center_id}>"

    def __hash__(self):
        return hash(self.demand_center_id)

    def __eq__(self, other) -> bool:
        return self.demand_center_id == other.demand_center_id


class SteelAllocations:
    def __init__(self, allocations: dict[tuple[Plant, FurnaceGroup, DemandCenter], Volumes]) -> None:
        self.allocations = allocations

    def __repr__(self) -> str:
        return f"{len(self.allocations)} SteelAllocations"

    def get_total(self) -> Volumes:
        return sum(self.allocations.values(), Volumes(0))


class Environment:
    """
    Class to track system environment, e.g. the collective macro-scale of the system
    such as total capacity of technologies to produce commodity X, cost curve of steel and
    iron per unit of production, prediction of market prices, emissions, capex, and optimal plant technology
    contains yearly demand dicts and the iteration year
    """

    def __init__(self, cost_of_x_csv: Path = Auto):
        """
        Initialise the environment with the capex values and switching capex
        (i.e. the capex needed switching from tech A --> B)
        """
        from .Capex import FULL_GREENFIELD_CAPEX, TECH_SWITCH_CAPEX

        self.year = Year(SIMULATION_START_YEAR)

        self.name_to_capex = {
            "greenfield": FULL_GREENFIELD_CAPEX.copy(),
            "brownfield": {k: v / 2 for k, v in FULL_GREENFIELD_CAPEX.items()},
        }
        # Initialize these as defaultdict instead of None
        self.steel_init_capacity: defaultdict[str, defaultdict[str, int]] = defaultdict(lambda: defaultdict(int))
        self.iron_init_capacity: defaultdict[str, defaultdict[str, int]] = defaultdict(lambda: defaultdict(int))
        self.technology_switch_capex = TECH_SWITCH_CAPEX.copy()

        self.cost_of_capital = self.initiate_industrial_asset_cost_of_capital(cost_of_x_csv=cost_of_x_csv)

    def initiate_industrial_asset_cost_of_capital(self, cost_of_x_csv: Path = Auto) -> dict[str, float]:
        """
        Load the cost of capital for industrial assets from cost_of_x file
        Returns: a dictionary with the cost of capital for each country
        """
        if not cost_of_x_csv:
            from ..config import settings

            cost_of_x_csv = settings.cost_of_x_csv

        with open(cost_of_x_csv, "r") as file:
            data = json.load(file)
        return dict(zip(data["Country code"].values(), data["Cost of equity - industrial assets"].values()))

    def _generate_cost_dict(self, world_furnace_groups: list[FurnaceGroup]) -> dict[str, dict[str, float]]:
        """
        Generate a dict representation of the costs for all furnace_groups
        """
        assets = {}
        for fg in world_furnace_groups:
            if fg.status.lower() not in ACTIVE_STATUSES or fg.capacity == 0:
                continue
            unit_cost = fg.unit_production_cost

            # Use furnace group ID as key
            assets[fg.furnace_group_id] = {"capacity": fg.capacity, "unit_cost_of_production": unit_cost}

        # Sort by unit cost of production
        self.cost_dict = assets
        return assets

    def generate_cost_curve(self, world_furnace_groups: list[FurnaceGroup]) -> list[dict]:
        """
        Generate a cost curve based on the unit cost of production for furnace groups.

        1) Generates and sorts the internal cost dictionary by the unit cost of production,
        2) Computes a cumulative production capacity and associates it with the corresponding production cost.


        Returns:
            list[dict]: A list of dictionaries representing the cost curve. Each dictionary has keys:
                - "cumulative_capacity" (float): The cumulative production capacity.
                - "production_cost" (float): The unit cost of production at that cumulative capacity.

        TODO: Add separation of steel and iron price.
        """
        self._generate_cost_dict(world_furnace_groups)
        sorted_assets = dict(sorted(self.cost_dict.items(), key=lambda item: item[1]["unit_cost_of_production"]))

        # Sort by unit cost of production
        # Generate cost curve
        cumulative_production: float = 0
        curve = []
        for fg_id, values in sorted_assets.items():
            cap = values["capacity"]
            cost = values["unit_cost_of_production"]

            cumulative_production += cap

            curve.append({"cumulative_capacity": cumulative_production, "production_cost": cost})
        self.steel_cost_curve = curve
        return curve

    def update_cost_curve(self, world_furnace_groups: list[FurnaceGroup], product_type="steel") -> list[dict]:
        """
        Update the cost curve for a given product type.
        Filters the furnace groups by product type and generates the cost curve.

        Parameters:
            world_furnace_groups (list[FurnaceGroup]): List of furnace groups.
            product_type (str, optional): Product type to filter by (default "steel").

        """
        return self.generate_cost_curve(
            [fg for fg in world_furnace_groups if fg.technology.product.lower() == product_type]
        )

    def extract_price_from_costcurve(self, demand: float) -> float:
        """
        Return the production cost for the first cumulative capacity that meets or exceeds the demand.

        Raises:
            ValueError: If the cost curve is empty or if demand exceeds maximum cumulative production.
        """
        # Handle empty cost curve
        if not self.steel_cost_curve:
            raise ValueError("Cost curve is empty")

        # Get the last entry (highest capacity)
        last_entry = self.steel_cost_curve[-1]

        # If demand exceeds available capacity, raise an error
        if demand > last_entry["cumulative_capacity"]:
            raise ValueError(f"Demand of {demand} exceeds maximum cumulative production")

        # Normal case - find first entry that meets or exceeds demand
        for entry in self.steel_cost_curve:
            if entry["cumulative_capacity"] >= demand:
                return entry["production_cost"]

        raise ValueError("capacity should always be greater than demand")  # This should never happen

    def _predict_new_market_price(self, new_furnace_group: FurnaceGroup, demand: float) -> float:
        """
        Takes in a list of furnace groups and predicts the new market price based on the cost curve, but doesn't store it.
        Returns:  The production cost at or above the demand level.
        """
        extended_list_of_furnace_groups = self.cost_dict.copy()
        extended_list_of_furnace_groups[new_furnace_group.furnace_group_id] = {
            "capacity": new_furnace_group.capacity,
            "unit_cost_of_production": new_furnace_group.unit_production_cost,
        }
        sorted_assets = dict(
            sorted(extended_list_of_furnace_groups.items(), key=lambda item: item[1]["unit_cost_of_production"])
        )

        # Sort by unit cost of production

        # Generate cost curve
        cumulative_capacity: float = 0
        for fg_id, values in sorted_assets.items():
            cap = values["capacity"]
            cost = values["unit_cost_of_production"]

            cumulative_capacity += cap

            if cumulative_capacity >= demand:
                return cost
        raise ValueError(f"Demand of {demand} exceeds maximum cumulative capacity of {cumulative_capacity}")

    def new_furnace_npv(
        self,
        cost_of_debt=0.05,
        equity_share=0.2,
        lifetime=20,
        expected_utilisation_rate=0.7,
        mvp: bool = True,
        demand: int = Auto,
    ):
        """
        Derives the most benefitial technology and its NPV for a new furnace

        Uses the calculate_npv_full function to calculate the NPV for each technology, and returns the highest NPV and the associated technology.
        """
        from steelo.domain import calculate_costs as cc
        from steelo.domain.Capex import FULL_GREENFIELD_CAPEX as greenfield_capex

        if not demand or demand is None:  # Auto or None? - this is probably wrong FIXME - jochen
            demand = int(self.current_demand)

        NPV = {}
        if mvp:
            for tech, capex in greenfield_capex.items():
                if tech in ["BOF", "DRI-EAF", "EAF"]:
                    NPV[tech] = cc.calculate_npv_full(
                        technology=tech,
                        capex=capex,
                        cost_of_debt=cost_of_debt,
                        equity_share=equity_share,
                        lifetime=lifetime,
                        fixed_opex=capex * 0.1,
                        expected_utilisation_rate=expected_utilisation_rate,
                        capacity=Volumes(1000),
                        price=self.extract_price_from_costcurve(demand),
                    )

        else:  # for now this is how we hack the mvp
            for tech, capex in greenfield_capex.items():
                NPV[tech] = cc.calculate_npv_full(
                    technology=tech,
                    capex=capex,
                    cost_of_debt=cost_of_debt,
                    equity_share=equity_share,
                    lifetime=lifetime,
                    fixed_opex=capex * 0.1,
                    expected_utilisation_rate=expected_utilisation_rate,
                    capacity=Volumes(1000),
                    price=self.extract_price_from_costcurve(demand),
                )
        # Find technology with maximum NPV value
        best = max(NPV, key=lambda k: NPV[k] if NPV[k] is not None else float("-inf"))
        best_value = NPV[best]
        return best_value, best

    def update_regional_capacity(self, world_plants: list[Plant]) -> None:
        """
        Extracts the regional capacity of the world's furnace groups depending on technology.
        """

        self.regional_steel_capacity: defaultdict[str, defaultdict[str, int]] = defaultdict(lambda: defaultdict(int))
        self.regional_iron_capacity: defaultdict[str, defaultdict[str, int]] = defaultdict(lambda: defaultdict(int))

        for pa in world_plants:
            for fg in pa.furnace_groups:
                if fg.status.lower() not in ACTIVE_STATUSES:
                    continue

                target_capacity = (
                    self.regional_steel_capacity
                    if fg.technology.product == "Steel"
                    else self.regional_iron_capacity
                    if fg.technology.product == "Iron"
                    else None
                )

                if target_capacity is not None and pa.location.iso3 is not None:
                    target_capacity[pa.location.iso3][fg.technology.name] += fg.capacity

        if not self.steel_init_capacity:
            self.steel_init_capacity = self.regional_steel_capacity.copy()

        if not self.iron_init_capacity:
            self.iron_init_capacity = self.regional_iron_capacity.copy()

        # TODO: fix division by zero error and uncomment
        # # Update capex reduction ratios if regional capacities have changed.
        # if self.steel_C0 != self.regional_steel_capacity:
        #     self.update_steel_capex_reduction_ratio()
        # if self.iron_C0 != self.regional_iron_capacity:
        #     self.update_iron_capex_reduction_ratio()

    def update_steel_capex_reduction_ratio(self) -> None:
        """
        Will calculate the ratio to reduce the capex of iron technologies with
        """
        from .calculate_costs import calculate_capex_reduction_rate

        self.steel_capex_reduction_ratio: dict[str, dict[str, float]] = {}
        for region, capacity in self.regional_steel_capacity.items():
            self.steel_capex_reduction_ratio[region] = {}
            for tech, cap in capacity.items():
                self.steel_capex_reduction_ratio[region][tech] = calculate_capex_reduction_rate(
                    self.steel_init_capacity[region][tech], cap
                )

    def update_iron_capex_reduction_ratio(self) -> None:
        """
        Will calculate the ratio to reduce the capex of iron technologies with
        """
        from .calculate_costs import calculate_capex_reduction_rate

        self.iron_capex_reduction_ratio: dict[str, dict[str, float]] = {}
        for region, capacity in self.regional_iron_capacity.items():
            self.iron_capex_reduction_ratio[region] = {}

            for tech, cap in capacity.items():
                self.iron_capex_reduction_ratio[region][tech] = calculate_capex_reduction_rate(
                    capacity_current=cap, capacity_zero=self.iron_init_capacity[region][tech]
                )
        return None

    def initiate_demand_dicts(self, world_demand_centres: list[DemandCenter]) -> None:
        """
        Initiates the demand dictionary in the environment

        This allows us to have a proxy for the demand centres, as values, but not locations, change in time
        """
        self.demand_dict: dict[str, dict[Year, Volumes]] = defaultdict(lambda: defaultdict(lambda: Volumes(0)))
        for dc in world_demand_centres:
            self.demand_dict[dc.demand_center_id] = dc.demand_by_year

    def calculate_demand(self) -> None:
        """
        Calaculates the total demand for the current year
        """
        self.current_demand = sum(entry.get(self.year, 0) for entry in self.demand_dict.values())



from .interface import EconomicModel
from .dummy import PlantsGoOutOfBusinessOverTime
from .plant_agent import PlantAgentsModel, AllocationModel

__all__ = ["EconomicModel", "PlantAgentsModel", "AllocationModel", "PlantsGoOutOfBusinessOverTime"]



from pathlib import Path
from typing import Iterable
import json
from datetime import date
from steelo.domain.models import TimeFrame

from .interface import PlantRepository, FurnaceGroupRepository, DemandCenterRepository
from ...domain import Plant, Volumes, Year, Location, Technology, FurnaceGroup, PointInTime, DemandCenter


class PlantInMemoryRepository:
    def __init__(self) -> None:
        self.data: dict[str, Plant] = {}
        self.furnace_id_to_plant_id: dict[str, str] = {}
        self.seen: set[Plant] = set()

    def add(self, plant: Plant) -> None:
        self.data[plant.plant_id] = plant
        for furnace_group in plant.furnace_groups:
            self.furnace_id_to_plant_id[furnace_group.furnace_group_id] = plant.plant_id
        self.seen.add(plant)

    def add_list(self, plants: Iterable[Plant]) -> None:
        for plant in plants:
            self.add(plant)

    def get(self, plant_id) -> Plant:
        return self.data[plant_id]

    def list(self) -> list[Plant]:
        return list(self.data.values())

    def get_by_furnace_group_id(self, furnace_group_id: str) -> Plant:
        plant_id = self.furnace_id_to_plant_id[furnace_group_id]
        return self.data[plant_id]


class FurnaceGroupInMemoryRepository:
    def __init__(self) -> None:
        self.data: dict[str, FurnaceGroup] = {}

    def add(self, furnace_group: FurnaceGroup) -> None:
        self.data[furnace_group.furnace_group_id] = furnace_group

    def add_list(self, furnace_groups: Iterable[FurnaceGroup]) -> None:
        for furnace_group in furnace_groups:
            self.add(furnace_group)

    def get(self, furnace_group_id) -> FurnaceGroup:
        return self.data[furnace_group_id]

    def list(self) -> list[FurnaceGroup]:
        return list(self.data.values())


class DemandCenterInMemoryRepository:
    def __init__(self) -> None:
        self.data: dict[str, DemandCenter] = {}

    def add(self, demand_center: DemandCenter) -> None:
        self.data[demand_center.demand_center_id] = demand_center

    def add_list(self, demand_centers: Iterable[DemandCenter]) -> None:
        for demand_center in demand_centers:
            self.add(demand_center)

    def get(self, demand_center_id) -> DemandCenter:
        return self.data[demand_center_id]

    def list(self) -> list[DemandCenter]:
        return list(self.data.values())


class InMemoryRepository:
    plants: PlantRepository
    furnace_groups: FurnaceGroupRepository
    demand_centers: DemandCenterRepository

    def __init__(self) -> None:
        self.plants = PlantInMemoryRepository()
        self.furnace_groups = FurnaceGroupInMemoryRepository()
        self.demand_centers = DemandCenterInMemoryRepository()


def import_plants_from_json(json_path: Path, mvp_filter_furnaces: bool = True) -> InMemoryRepository:
    """
    Import plants data from a JSON file and add it to an in-memory repository.
    """
    with json_path.open("r", encoding="utf-8") as f:
        plants_data = json.load(f)

    def convert_to_domain(plant_dict):
        # Convert the raw dictionary to a domain Plant object
        location = Location(**plant_dict["location"])
        furnace_groups = []
        for fg_dict in plant_dict["furnace_groups"]:
            technology = Technology(**fg_dict["technology"])
            time_frame = TimeFrame(**fg_dict["lifetime"]["time_frame"])
            lifetime = PointInTime(current=Year(fg_dict["lifetime"]["current"]), time_frame=time_frame)
            furnace_group = FurnaceGroup(
                furnace_group_id=fg_dict["furnace_group_id"],
                capacity=Volumes(fg_dict["capacity"]),
                status=fg_dict["status"],
                lifetime=lifetime,
                last_renovation_date=date.fromisoformat(fg_dict["last_renovation_date"]),
                technology=technology,
                historical_production={Year(k): Volumes(v) for k, v in fg_dict["historical_production"].items()},
                utilization_rate=fg_dict["utilization_rate"],
            )
            furnace_groups.append(furnace_group)
        return Plant(
            plant_id=plant_dict["plant_id"],
            location=location,
            furnace_groups=(
                [fg for fg in furnace_groups if fg.technology.name in ["EAF", "BOF"]]
                if mvp_filter_furnaces
                else furnace_groups
            ),
            power_source=plant_dict["power_source"],
            soe_status=plant_dict["soe_status"],
            parent_gem_id=plant_dict["parent_gem_id"],
            workforce_size=plant_dict["workforce_size"],
            certified=plant_dict["certified"],
            category_steel_product=set(plant_dict["category_steel_product"]),
        )

    plants = [convert_to_domain(plant_dict) for plant_dict in plants_data["root"]]

    print({fg.technology.name for plant in plants for fg in plant.furnace_groups})

    repository_in_memory = InMemoryRepository()

    repository_in_memory.plants.add_list(plants)
    return repository_in_memory



from dataclasses import dataclass


@dataclass(frozen=True)
class CountryData:
    """Represents all available data for a country."""

    country_name: str
    country_code: str
    irena_name: str
    region: str
    ssp_region: str
    gem_country: str | None = None
    ws_region: str | None = None


COUNTRY_MASTER_DATA: dict[str, CountryData] = {
    "Afghanistan": CountryData(
        country_name="Afghanistan",
        country_code="AFG",
        irena_name="Afghanistan",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Angola": CountryData(
        country_name="Angola",
        country_code="AGO",
        irena_name="Angola",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country="Angola",
        ws_region="Other Africa",
    ),
    "Albania": CountryData(
        country_name="Albania",
        country_code="ALB",
        irena_name="Albania",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="Albania",
        ws_region="Other Europe",
    ),
    "United Arab Emirates": CountryData(
        country_name="United Arab Emirates",
        country_code="ARE",
        irena_name="United Arab Emirates (the)",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="United Arab Emirates",
        ws_region="Other Middle East",
    ),
    "Argentina": CountryData(
        country_name="Argentina",
        country_code="ARG",
        irena_name="Argentina",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Argentina",
        ws_region="Argentina",
    ),
    "Armenia": CountryData(
        country_name="Armenia",
        country_code="ARM",
        irena_name="Armenia",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Australia": CountryData(
        country_name="Australia",
        country_code="AUS",
        irena_name="Australia",
        region="Oceania",
        ssp_region="R5.2OECD",
        gem_country="Australia",
        ws_region="Australia",
    ),
    "Austria": CountryData(
        country_name="Austria",
        country_code="AUT",
        irena_name="Austria",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Austria",
        ws_region="Austria",
    ),
    "Azerbaijan": CountryData(
        country_name="Azerbaijan",
        country_code="AZE",
        irena_name="Azerbaijan",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country="Azerbaijan",
        ws_region="Other CIS",
    ),
    "Burundi": CountryData(
        country_name="Burundi",
        country_code="BDI",
        irena_name="Burundi",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Belgium": CountryData(
        country_name="Belgium",
        country_code="BEL",
        irena_name="Belgium",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Belgium",
        ws_region="Belgium",
    ),
    "Benin": CountryData(
        country_name="Benin",
        country_code="BEN",
        irena_name="Benin",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Burkina Faso": CountryData(
        country_name="Burkina Faso",
        country_code="BFA",
        irena_name="Burkina Faso",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Bangladesh": CountryData(
        country_name="Bangladesh",
        country_code="BGD",
        irena_name="Bangladesh",
        region="India",
        ssp_region="R5.2ASIA",
        gem_country="Bangladesh",
        ws_region="Other Asia",
    ),
    "Bulgaria": CountryData(
        country_name="Bulgaria",
        country_code="BGR",
        irena_name="Bulgaria",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="Bulgaria",
        ws_region="Bulgaria",
    ),
    "Bahamas": CountryData(
        country_name="Bahamas",
        country_code="BHS",
        irena_name="Bahamas (the)",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Bahrain": CountryData(
        country_name="Bahrain",
        country_code="BHR",
        irena_name="Bahrain",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Bahrain",
        ws_region="Other Middle East",
    ),
    "Bosnia and Herz.": CountryData(
        country_name="Bosnia and Herz.",
        country_code="BIH",
        irena_name="Bosnia and Herzegovina",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="Bosnia and Herzegovina",
        ws_region="Other Europe",
    ),
    "Belarus": CountryData(
        country_name="Belarus",
        country_code="BLR",
        irena_name="Belarus",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country="Belarus",
        ws_region="Other CIS",
    ),
    "Belize": CountryData(
        country_name="Belize",
        country_code="BLZ",
        irena_name="Belize",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Bolivia": CountryData(
        country_name="Bolivia",
        country_code="BOL",
        irena_name="Bolivia (Plurinational State of)",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Bolivia",
        ws_region="Other Central & South America",
    ),
    "Brazil": CountryData(
        country_name="Brazil",
        country_code="BRA",
        irena_name="Brazil",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Brazil",
        ws_region="Brazil",
    ),
    "Brunei": CountryData(
        country_name="Brunei",
        country_code="BRN",
        irena_name="Brunei Darussalam",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Bhutan": CountryData(
        country_name="Bhutan",
        country_code="BTN",
        irena_name="Bhutan",
        region="India",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Botswana": CountryData(
        country_name="Botswana",
        country_code="BWA",
        irena_name="Botswana",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Central African Rep.": CountryData(
        country_name="Central African Rep.",
        country_code="CAF",
        irena_name="Central African Republic (the)",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Canada": CountryData(
        country_name="Canada",
        country_code="CAN",
        irena_name="Canada",
        region="North America",
        ssp_region="R5.2OECD",
        gem_country="Canada",
        ws_region="Canada",
    ),
    "Switzerland": CountryData(
        country_name="Switzerland",
        country_code="CHE",
        irena_name="Switzerland",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Switzerland",
        ws_region="Other Europe",
    ),
    "Chile": CountryData(
        country_name="Chile",
        country_code="CHL",
        irena_name="Chile",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Chile",
        ws_region="Chile",
    ),
    "China": CountryData(
        country_name="China",
        country_code="CHN",
        irena_name="China",
        region="China",
        ssp_region="R5.2ASIA",
        gem_country="China",
        ws_region="China",
    ),
    "Côte d'Ivoire": CountryData(
        country_name="Côte d'Ivoire",
        country_code="CIV",
        irena_name="Côte d'Ivoire",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Cameroon": CountryData(
        country_name="Cameroon",
        country_code="CMR",
        irena_name="Cameroon",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Dem. Rep. Congo": CountryData(
        country_name="Dem. Rep. Congo",
        country_code="COD",
        irena_name="Democratic Republic of the Congo (the)",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Congo": CountryData(
        country_name="Congo",
        country_code="COG",
        irena_name="Congo (the)",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Colombia": CountryData(
        country_name="Colombia",
        country_code="COL",
        irena_name="Colombia",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Costa Rica": CountryData(
        country_name="Costa Rica",
        country_code="CRI",
        irena_name="Costa Rica",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Cuba": CountryData(
        country_name="Cuba",
        country_code="CUB",
        irena_name="Cuba",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "N. Cyprus": CountryData(
        country_name="N. Cyprus",
        country_code="CYP",
        irena_name="Cyprus",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Cyprus": CountryData(
        country_name="Cyprus",
        country_code="CYP",
        irena_name="Cyprus",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Czechia": CountryData(
        country_name="Czechia",
        country_code="CZE",
        irena_name="Czechia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Czech Republic",
        ws_region="Czechia",
    ),
    "Germany": CountryData(
        country_name="Germany",
        country_code="DEU",
        irena_name="Germany",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Germany",
        ws_region="Germany",
    ),
    "Djibouti": CountryData(
        country_name="Djibouti",
        country_code="DJI",
        irena_name="Djibouti",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Denmark": CountryData(
        country_name="Denmark",
        country_code="DNK",
        irena_name="Denmark",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Dominican Rep.": CountryData(
        country_name="Dominican Rep.",
        country_code="DOM",
        irena_name="Dominican Republic (the)",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Algeria": CountryData(
        country_name="Algeria",
        country_code="DZA",
        irena_name="Algeria",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country="Algeria",
        ws_region="Other Africa",
    ),
    "Ecuador": CountryData(
        country_name="Ecuador",
        country_code="ECU",
        irena_name="Ecuador",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Egypt": CountryData(
        country_name="Egypt",
        country_code="EGY",
        irena_name="Egypt",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country="Egypt",
        ws_region="Egypt",
    ),
    "Eritrea": CountryData(
        country_name="Eritrea",
        country_code="ERI",
        irena_name="Eritrea",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Spain": CountryData(
        country_name="Spain",
        country_code="ESP",
        irena_name="Spain",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Spain",
        ws_region="Spain",
    ),
    "Estonia": CountryData(
        country_name="Estonia",
        country_code="EST",
        irena_name="Estonia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Ethiopia": CountryData(
        country_name="Ethiopia",
        country_code="ETH",
        irena_name="Ethiopia",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country="Ethiopia",
        ws_region="Other Africa",
    ),
    "Finland": CountryData(
        country_name="Finland",
        country_code="FIN",
        irena_name="Finland",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Finland",
        ws_region="Finland",
    ),
    "Fiji": CountryData(
        country_name="Fiji",
        country_code="FJI",
        irena_name="Fiji",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Falkland Is.": CountryData(
        country_name="Falkland Is.",
        country_code="FLK",
        irena_name="Falkland Islands (Malvinas)",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "France": CountryData(
        country_name="France",
        country_code="FRA",
        irena_name="France",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="France",
        ws_region="France",
    ),
    "Gabon": CountryData(
        country_name="Gabon",
        country_code="GAB",
        irena_name="Gabon",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "United Kingdom": CountryData(
        country_name="United Kingdom",
        country_code="GBR",
        irena_name="United Kingdom of Great Britain and Northern Ireland (the)",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="United Kingdom",
        ws_region="United Kingdom",
    ),
    "Georgia": CountryData(
        country_name="Georgia",
        country_code="GEO",
        irena_name="Georgia",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Ghana": CountryData(
        country_name="Ghana",
        country_code="GHA",
        irena_name="Ghana",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country="Ghana",
        ws_region="Other Africa",
    ),
    "Guinea": CountryData(
        country_name="Guinea",
        country_code="GIN",
        irena_name="Guinea",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Gambia": CountryData(
        country_name="Gambia",
        country_code="GMB",
        irena_name="Gambia (the)",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Guinea-Bissau": CountryData(
        country_name="Guinea-Bissau",
        country_code="GNB",
        irena_name="Guinea-Bissau",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Eq. Guinea": CountryData(
        country_name="Eq. Guinea",
        country_code="GNQ",
        irena_name="Guinea",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Greece": CountryData(
        country_name="Greece",
        country_code="GRC",
        irena_name="Greece",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Greece",
        ws_region="Greece",
    ),
    "Greenland": CountryData(
        country_name="Greenland",
        country_code="GRL",
        irena_name="Greenland",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Guatemala": CountryData(
        country_name="Guatemala",
        country_code="GTM",
        irena_name="Guatemala",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country="Guatemala",
        ws_region="Other Central & South America",
    ),
    "Guyana": CountryData(
        country_name="Guyana",
        country_code="GUY",
        irena_name="Guyana",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Honduras": CountryData(
        country_name="Honduras",
        country_code="HND",
        irena_name="Honduras",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Croatia": CountryData(
        country_name="Croatia",
        country_code="HRV",
        irena_name="Croatia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Haiti": CountryData(
        country_name="Haiti",
        country_code="HTI",
        irena_name="Haiti",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Hong Kong": CountryData(
        country_name="Hong Kong",
        country_code="HKG",
        irena_name="Hong Kong",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Hong Kong",
        ws_region="China & Taiwan",
    ),
    "Hungary": CountryData(
        country_name="Hungary",
        country_code="HUN",
        irena_name="Hungary",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Hungary",
        ws_region="Hungary",
    ),
    "Indonesia": CountryData(
        country_name="Indonesia",
        country_code="IDN",
        irena_name="Indonesia",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Indonesia",
        ws_region="Other Asia",
    ),
    "India": CountryData(
        country_name="India",
        country_code="IND",
        irena_name="India",
        region="India",
        ssp_region="R5.2ASIA",
        gem_country="India",
        ws_region="India",
    ),
    "Ireland": CountryData(
        country_name="Ireland",
        country_code="IRL",
        irena_name="Ireland",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Iran": CountryData(
        country_name="Iran",
        country_code="IRN",
        irena_name="Iran (Islamic Republic of)",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Iran",
        ws_region="Iran",
    ),
    "Iraq": CountryData(
        country_name="Iraq",
        country_code="IRQ",
        irena_name="Iraq",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Iraq",
        ws_region="Other Middle East",
    ),
    "Iceland": CountryData(
        country_name="Iceland",
        country_code="ISL",
        irena_name="Iceland",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Israel": CountryData(
        country_name="Israel",
        country_code="ISR",
        irena_name="Israel",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Italy": CountryData(
        country_name="Italy",
        country_code="ITA",
        irena_name="Italy",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Italy",
        ws_region="Italy",
    ),
    "Jamaica": CountryData(
        country_name="Jamaica",
        country_code="JAM",
        irena_name="Jamaica",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Jordan": CountryData(
        country_name="Jordan",
        country_code="JOR",
        irena_name="Jordan",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Japan": CountryData(
        country_name="Japan",
        country_code="JPN",
        irena_name="Japan",
        region="South + East Asia",
        ssp_region="R5.2OECD",
        gem_country="Japan",
        ws_region="Japan",
    ),
    "Kazakhstan": CountryData(
        country_name="Kazakhstan",
        country_code="KAZ",
        irena_name="Kazakhstan",
        region="Russia + CIS",
        ssp_region="R5.2REF",
        gem_country="Kazakhstan",
        ws_region="Other CIS",
    ),
    "Kenya": CountryData(
        country_name="Kenya",
        country_code="KEN",
        irena_name="Kenya",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country="Kenya",
        ws_region="Other Africa",
    ),
    "Kyrgyzstan": CountryData(
        country_name="Kyrgyzstan",
        country_code="KGZ",
        irena_name="Kyrgyzstan",
        region="Russia + CIS",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Cambodia": CountryData(
        country_name="Cambodia",
        country_code="KHM",
        irena_name="Cambodia",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Cambodia",
        ws_region="Other Asia",
    ),
    "South Korea": CountryData(
        country_name="South Korea",
        country_code="KOR",
        irena_name="Democratic People's Republic of Korea (the)",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="South Korea",
        ws_region="South Korea",
    ),
    "Kosovo": CountryData(
        country_name="Kosovo",
        country_code="KOS",
        irena_name="Kosovo",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Kuwait": CountryData(
        country_name="Kuwait",
        country_code="KWT",
        irena_name="Kuwait",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Kuwait",
        ws_region="Other Middle East",
    ),
    "Laos": CountryData(
        country_name="Laos",
        country_code="LAO",
        irena_name="Lao People's Democratic Republic (the)",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Lebanon": CountryData(
        country_name="Lebanon",
        country_code="LBN",
        irena_name="Lebanon",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Liberia": CountryData(
        country_name="Liberia",
        country_code="LBR",
        irena_name="Liberia",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Libya": CountryData(
        country_name="Libya",
        country_code="LBY",
        irena_name="Libya",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country="Libya",
        ws_region="Other Africa",
    ),
    "Sri Lanka": CountryData(
        country_name="Sri Lanka",
        country_code="LKA",
        irena_name="Sri Lanka",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Sri Lanka",
        ws_region="Other Asia",
    ),
    "Lesotho": CountryData(
        country_name="Lesotho",
        country_code="LSO",
        irena_name="Lesotho",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Lithuania": CountryData(
        country_name="Lithuania",
        country_code="LTU",
        irena_name="Lithuania",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Luxembourg": CountryData(
        country_name="Luxembourg",
        country_code="LUX",
        irena_name="Luxembourg",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Luxembourg",
        ws_region="Luxembourg",
    ),
    "Latvia": CountryData(
        country_name="Latvia",
        country_code="LVA",
        irena_name="Latvia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Latvia",
        ws_region="Other Europe",
    ),
    "Morocco": CountryData(
        country_name="Morocco",
        country_code="MAR",
        irena_name="Morocco",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country="Morocco",
        ws_region="Other Africa",
    ),
    "Moldova": CountryData(
        country_name="Moldova",
        country_code="MDA",
        irena_name="Republic of Moldova (the)",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country="Moldova",
        ws_region="Other CIS",
    ),
    "Madagascar": CountryData(
        country_name="Madagascar",
        country_code="MDG",
        irena_name="Madagascar",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Mexico": CountryData(
        country_name="Mexico",
        country_code="MEX",
        irena_name="Mexico",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country="Mexico",
        ws_region="Mexico",
    ),
    "North Macedonia": CountryData(
        country_name="North Macedonia",
        country_code="MKD",
        irena_name="North Macedonia",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="North Macedonia",
        ws_region="Other Europe",
    ),
    "Mali": CountryData(
        country_name="Mali",
        country_code="MLI",
        irena_name="Mali",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Myanmar": CountryData(
        country_name="Myanmar",
        country_code="MMR",
        irena_name="Myanmar",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Myanmar",
        ws_region="Other Asia",
    ),
    "Montenegro": CountryData(
        country_name="Montenegro",
        country_code="MNE",
        irena_name="Montenegro",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country=None,
        ws_region=None,
    ),
    "Mongolia": CountryData(
        country_name="Mongolia",
        country_code="MNG",
        irena_name="Mongolia",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Mozambique": CountryData(
        country_name="Mozambique",
        country_code="MOZ",
        irena_name="Mozambique",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country="Mozambique",
        ws_region="Other Africa",
    ),
    "Mauritania": CountryData(
        country_name="Mauritania",
        country_code="MRT",
        irena_name="Mauritania",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Malawi": CountryData(
        country_name="Malawi",
        country_code="MWI",
        irena_name="Malawi",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Malaysia": CountryData(
        country_name="Malaysia",
        country_code="MYS",
        irena_name="Malaysia",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Malaysia",
        ws_region="Other Asia",
    ),
    "Namibia": CountryData(
        country_name="Namibia",
        country_code="NAM",
        irena_name="Namibia",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country="Namibia",
        ws_region="Other Africa",
    ),
    "New Caledonia": CountryData(
        country_name="New Caledonia",
        country_code="NCL",
        irena_name="New Caledonia",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Niger": CountryData(
        country_name="Niger",
        country_code="NER",
        irena_name="Niger (the)",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Nigeria": CountryData(
        country_name="Nigeria",
        country_code="NGA",
        irena_name="Nigeria",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country="Nigeria",
        ws_region="Other Africa",
    ),
    "Nicaragua": CountryData(
        country_name="Nicaragua",
        country_code="NIC",
        irena_name="Nicaragua",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Netherlands": CountryData(
        country_name="Netherlands",
        country_code="NLD",
        irena_name="Netherlands (Kingdom of the)",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Netherlands",
        ws_region="Netherlands",
    ),
    "Norway": CountryData(
        country_name="Norway",
        country_code="NOR",
        irena_name="Norway",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Norway",
        ws_region="Other Europe",
    ),
    "Nepal": CountryData(
        country_name="Nepal",
        country_code="NPL",
        irena_name="Nepal",
        region="India",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "New Zealand": CountryData(
        country_name="New Zealand",
        country_code="NZL",
        irena_name="New Zealand",
        region="Oceania",
        ssp_region="R5.2OECD",
        gem_country="New Zealand",
        ws_region="New Zealand",
    ),
    "Oman": CountryData(
        country_name="Oman",
        country_code="OMN",
        irena_name="Oman",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Oman",
        ws_region="Other Middle East",
    ),
    "Pakistan": CountryData(
        country_name="Pakistan",
        country_code="PAK",
        irena_name="Pakistan",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Pakistan",
        ws_region="Other Asia",
    ),
    "Panama": CountryData(
        country_name="Panama",
        country_code="PAN",
        irena_name="Panama",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Peru": CountryData(
        country_name="Peru",
        country_code="PER",
        irena_name="Peru",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Peru",
        ws_region="Other Central & South America",
    ),
    "Philippines": CountryData(
        country_name="Philippines",
        country_code="PHL",
        irena_name="Philippines (the)",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Philippines",
        ws_region="Other Asia",
    ),
    "Papua New Guinea": CountryData(
        country_name="Papua New Guinea",
        country_code="PNG",
        irena_name="Papua New Guinea",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Poland": CountryData(
        country_name="Poland",
        country_code="POL",
        irena_name="Poland",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Poland",
        ws_region="Poland",
    ),
    "Puerto Rico": CountryData(
        country_name="Puerto Rico",
        country_code="PRI",
        irena_name="Puerto Rico",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "North Korea": CountryData(
        country_name="North Korea",
        country_code="PRK",
        irena_name="Republic of Korea (the)",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="North Korea",
        ws_region="Other Asia",
    ),
    "Portugal": CountryData(
        country_name="Portugal",
        country_code="PRT",
        irena_name="Portugal",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Portugal",
        ws_region="Portugal",
    ),
    "Paraguay": CountryData(
        country_name="Paraguay",
        country_code="PRY",
        irena_name="Paraguay",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Palestine": CountryData(
        country_name="Palestine",
        country_code="PSE",
        irena_name="State of Palestine (the)",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Qatar": CountryData(
        country_name="Qatar",
        country_code="QAT",
        irena_name="Qatar",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Qatar",
        ws_region="Other Middle East",
    ),
    "Romania": CountryData(
        country_name="Romania",
        country_code="ROU",
        irena_name="Romania",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Romania",
        ws_region="Romania",
    ),
    "Russia": CountryData(
        country_name="Russia",
        country_code="RUS",
        irena_name="Russian Federation (the)",
        region="Russia + CIS",
        ssp_region="R5.2REF",
        gem_country="Russia",
        ws_region="Russia",
    ),
    "Rwanda": CountryData(
        country_name="Rwanda",
        country_code="RWA",
        irena_name="Rwanda",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Saudi Arabia": CountryData(
        country_name="Saudi Arabia",
        country_code="SAU",
        irena_name="Saudi Arabia",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Saudi Arabia",
        ws_region="Saudi Arabia",
    ),
    "Sudan": CountryData(
        country_name="Sudan",
        country_code="SDN",
        irena_name="Sudan (the)",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Senegal": CountryData(
        country_name="Senegal",
        country_code="SEN",
        irena_name="Senegal",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Singapore": CountryData(
        country_name="Singapore",
        country_code="SGP",
        irena_name="Singapore",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Singapore",
        ws_region="Other Asia",
    ),
    "Solomon Is.": CountryData(
        country_name="Solomon Is.",
        country_code="SLB",
        irena_name="Solomon Islands",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Sierra Leone": CountryData(
        country_name="Sierra Leone",
        country_code="SLE",
        irena_name="Sierra Leone",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "El Salvador": CountryData(
        country_name="El Salvador",
        country_code="SLV",
        irena_name="El Salvador",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Somalia": CountryData(
        country_name="Somalia",
        country_code="SOM",
        irena_name="Somalia",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Somaliland": CountryData(
        country_name="Somaliland",
        country_code="SOM",
        irena_name="Somalia",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Serbia": CountryData(
        country_name="Serbia",
        country_code="SRB",
        irena_name="Serbia",
        region="Other Europe",
        ssp_region="R5.2OECD",
        gem_country="Serbia",
        ws_region="Other Europe",
    ),
    "S. Sudan": CountryData(
        country_name="S. Sudan",
        country_code="SSD",
        irena_name="South Sudan",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Suriname": CountryData(
        country_name="Suriname",
        country_code="SUR",
        irena_name="Suriname",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "Slovakia": CountryData(
        country_name="Slovakia",
        country_code="SVK",
        irena_name="Slovakia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Slovakia",
        ws_region="Slovakia",
    ),
    "Slovenia": CountryData(
        country_name="Slovenia",
        country_code="SVN",
        irena_name="Slovenia",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Slovenia",
        ws_region="Slovenia",
    ),
    "Sweden": CountryData(
        country_name="Sweden",
        country_code="SWE",
        irena_name="Sweden",
        region="EU + Schengen",
        ssp_region="R5.2OECD",
        gem_country="Sweden",
        ws_region="Sweden",
    ),
    "eSwatini": CountryData(
        country_name="eSwatini",
        country_code="SWZ",
        irena_name="Eswatini",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Syria": CountryData(
        country_name="Syria",
        country_code="SYR",
        irena_name="Syrian Arab Republic (the)",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country="Syria",
        ws_region="Other Middle East",
    ),
    "Chad": CountryData(
        country_name="Chad",
        country_code="TCD",
        irena_name="Chad",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Togo": CountryData(
        country_name="Togo",
        country_code="TGO",
        irena_name="Togo",
        region="West Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Thailand": CountryData(
        country_name="Thailand",
        country_code="THA",
        irena_name="Thailand",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Thailand",
        ws_region="Other Asia",
    ),
    "Tajikistan": CountryData(
        country_name="Tajikistan",
        country_code="TJK",
        irena_name="Tajikistan",
        region="Russia + CIS",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Turkmenistan": CountryData(
        country_name="Turkmenistan",
        country_code="TKM",
        irena_name="Turkmenistan",
        region="Russia + CIS",
        ssp_region="R5.2REF",
        gem_country=None,
        ws_region=None,
    ),
    "Timor-Leste": CountryData(
        country_name="Timor-Leste",
        country_code="TLS",
        irena_name="Timor-Leste",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "Trinidad and Tobago": CountryData(
        country_name="Trinidad and Tobago",
        country_code="TTO",
        irena_name="Trinidad and Tobago",
        region="Central America",
        ssp_region="R5.2LAM",
        gem_country="Trinidad and Tobago",
        ws_region="Other North America",
    ),
    "Tunisia": CountryData(
        country_name="Tunisia",
        country_code="TUN",
        irena_name="Tunisia",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Turkey": CountryData(
        country_name="Turkey",
        country_code="TUR",
        irena_name="Türkiye",
        region="Middle East",
        ssp_region="R5.2OECD",
        gem_country="Türkiye",
        ws_region="Turkey",
    ),
    "Taiwan": CountryData(
        country_name="Taiwan",
        country_code="TWN",
        irena_name="Chinese Taipei",
        region="China",
        ssp_region="R5.2ASIA",
        gem_country="Taiwan",
        ws_region="China & Taiwan",
    ),
    "Tanzania": CountryData(
        country_name="Tanzania",
        country_code="TZA",
        irena_name="United Republic of Tanzania (the)",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Uganda": CountryData(
        country_name="Uganda",
        country_code="UGA",
        irena_name="Uganda",
        region="East Africa",
        ssp_region="R5.2MAF",
        gem_country="Uganda",
        ws_region="Other Africa",
    ),
    "Ukraine": CountryData(
        country_name="Ukraine",
        country_code="UKR",
        irena_name="Ukraine",
        region="Other Europe",
        ssp_region="R5.2REF",
        gem_country="Ukraine",
        ws_region="Ukraine",
    ),
    "Uruguay": CountryData(
        country_name="Uruguay",
        country_code="URY",
        irena_name="Uruguay",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country=None,
        ws_region=None,
    ),
    "United States of America": CountryData(
        country_name="United States of America",
        country_code="USA",
        irena_name="United States of America (the)",
        region="North America",
        ssp_region="R5.2OECD",
        gem_country="United States",
        ws_region="United States",
    ),
    "Uzbekistan": CountryData(
        country_name="Uzbekistan",
        country_code="UZB",
        irena_name="Uzbekistan",
        region="Russia + CIS",
        ssp_region="R5.2ASIA",
        gem_country="Uzbekistan",
        ws_region="Other CIS",
    ),
    "Venezuela": CountryData(
        country_name="Venezuela",
        country_code="VEN",
        irena_name="Venezuela (Bolivarian Republic of)",
        region="South America",
        ssp_region="R5.2LAM",
        gem_country="Venezuela",
        ws_region="Venezuela",
    ),
    "Vietnam": CountryData(
        country_name="Vietnam",
        country_code="VNM",
        irena_name="Viet Nam",
        region="South + East Asia",
        ssp_region="R5.2ASIA",
        gem_country="Vietnam",
        ws_region="Other Asia",
    ),
    "Vanuatu": CountryData(
        country_name="Vanuatu",
        country_code="VUT",
        irena_name="Vanuatu",
        region="Oceania",
        ssp_region="R5.2ASIA",
        gem_country=None,
        ws_region=None,
    ),
    "W. Sahara": CountryData(
        country_name="W. Sahara",
        country_code="WES",
        irena_name="Morocco",
        region="North Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Yemen": CountryData(
        country_name="Yemen",
        country_code="YEM",
        irena_name="Yemen",
        region="Middle East",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "South Africa": CountryData(
        country_name="South Africa",
        country_code="ZAF",
        irena_name="South Africa",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country="South Africa",
        ws_region="South Africa",
    ),
    "Zambia": CountryData(
        country_name="Zambia",
        country_code="ZMB",
        irena_name="Zambia",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country=None,
        ws_region=None,
    ),
    "Zimbabwe": CountryData(
        country_name="Zimbabwe",
        country_code="ZWE",
        irena_name="Zimbabwe",
        region="Southern Africa",
        ssp_region="R5.2MAF",
        gem_country="Zimbabwe",
        ws_region="Other Africa",
    ),
    # Add any additional countries here if needed
}

COUNTRY_TO_CODE_MAP: dict[str, str] = {country: data.country_code for country, data in COUNTRY_MASTER_DATA.items()}

COUNTRY_TO_IRENA_MAP: dict[str, str] = {country: data.irena_name for country, data in COUNTRY_MASTER_DATA.items()}

COUNTRY_TO_REGION_MAP: dict[str, str] = {country: data.region for country, data in COUNTRY_MASTER_DATA.items()}

COUNTRY_TO_SSP_REGION_MAP: dict[str, str] = {country: data.ssp_region for country, data in COUNTRY_MASTER_DATA.items()}

# For GEM country map, only include entries where gem_country is not None
GEM_COUNTRY_WS_REGION_MAP: dict[str, str] = {
    data.gem_country: data.ws_region
    for data in COUNTRY_MASTER_DATA.values()
    if data.gem_country is not None and data.ws_region is not None
}



from .models import (
    Environment,
    FurnaceGroup,
    Location,
    Plant,
    Technology,
    Volumes,
    Year,
    PointInTime,
    ProductCategory,
    TimeFrame,
    DemandCenter,
    PrimaryFeedstock,
)
from .events import Event
from .commands import Command
from .constants import Auto

__all__ = [
    "Auto",
    "Command",
    "Environment",
    "Event",
    "FurnaceGroup",
    "Location",
    "PointInTime",
    "ProductCategory",
    "Plant",
    "Technology",
    "TimeFrame",
    "Volumes",
    "Year",
    "DemandCenter",
    "PrimaryFeedstock",
]



bom_data = {
    ("iron_bf", "Pellets_high", "Coke", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke", "Energy", "Coke", "Reductant"): {"Unit": "t/t product", "Value": 0.4445},
    ("iron_bf", "Pellets_high", "Coke", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 28.073579404523805,
    },
    ("iron_bf", "Pellets_high", "Coke", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.8302102199789223,
    },
    ("iron_bf", "Pellets_high", "Coke", "Energy", "Hydrogen", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.019299934640706815,
    },
    ("iron_bf", "Pellets_high", "Coke", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.047284839869731705,
    },
    ("iron_bf", "Pellets_high", "Coke", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.3875335131550575,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.2625832054586857,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 28.073579404523805,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.8302102199789223,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "Hydrogen", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke+PCI", "Energy", "PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.20212977171257143,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.019299934640706815,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.047284839869731705,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.3875335131550575,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.2625832054586857,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 28.073579404523805,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.8302102199789223,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.02573411445414683,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Energy", "PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.08983545409447617,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.019299934640706815,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.047284839869731705,
    },
    ("iron_bf", "Pellets_high", "Coke+PCI+H2", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.3875335131550575,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "Bio-PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.20212977171257143,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.2625832054586857,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 28.073579404523805,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.8302102199789223,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.0,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.019299934640706815,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.047284839869731705,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.3875335131550575,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "Bio-PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.08983545409447617,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.2625832054586857,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 28.073579404523805,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.8302102199789223,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.02573411445414683,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.019299934640706815,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.047284839869731705,
    },
    ("iron_bf", "Pellets_high", "Coke+bio-PCI+H2", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.3875335131550575,
    },
    ("iron_bf", "Pellets_high", "null", "Constraint", "Pellets_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "Pellets_high", "null", "Constraint", "Pellets_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.49999999999999994,
    },
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 29.999999999999996,
    },
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 2.2},
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "Hydrogen", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.0602373338317374,
    },
    ("iron_bf", "Pellets_mid", "Coke", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.14758146788775664,
    },
    ("iron_bf", "Pellets_mid", "Coke", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.482746635006723,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.3055999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 29.999999999999996,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 2.2},
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "Hydrogen", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.216},
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.0602373338317374,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.14758146788775664,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.482746635006723,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "Bio-PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.3055999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 29.999999999999996,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 2.2},
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.027500000000000004,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Energy", "PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.09599999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.0602373338317374,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.14758146788775664,
    },
    ("iron_bf", "Pellets_mid", "Coke+PCI+H2", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.482746635006723,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "Bio-PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.216,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.3055999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 29.999999999999996,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 2.2},
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.0,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.0602373338317374,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.14758146788775664,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.482746635006723,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "Bio-PCI", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.09599999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "Coke", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.3055999999999999,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 29.999999999999996,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 2.2},
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "t/t product",
        "Value": 0.027500000000000004,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Energy", "PCI", "Reductant"): {"Unit": "t/t product", "Value": 0.0},
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.0602373338317374,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.14758146788775664,
    },
    ("iron_bf", "Pellets_mid", "Coke+bio-PCI+H2", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.482746635006723,
    },
    ("iron_bf", "Pellets_mid", "null", "Constraint", "Pellets_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "Pellets_mid", "null", "Constraint", "Pellets_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "null", "Coke", "Constraint", "Coke", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "null", "Coke", "Constraint", "Coke", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "null", "Coke+PCI", "Constraint", "Coke+PCI", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "null", "Coke+PCI", "Constraint", "Coke+PCI", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "null", "Coke+PCI+H2", "Constraint", "Coke+PCI+H2", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "null", "Coke+PCI+H2", "Constraint", "Coke+PCI+H2", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "null", "Coke+bio-PCI", "Constraint", "Coke+bio-PCI", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "null", "Coke+bio-PCI", "Constraint", "Coke+bio-PCI", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_bf", "null", "Coke+bio-PCI+H2", "Constraint", "Coke+bio-PCI+H2", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_bf", "null", "Coke+bio-PCI+H2", "Constraint", "Coke+bio-PCI+H2", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_dri", "Pellets_high", "Coal", "Energy", "Coal", "Heat"): {"Unit": "GJ/t product", "Value": 0.0},
    ("iron_dri", "Pellets_high", "Coal", "Energy", "Coal", "Reductant"): {"Unit": "GJ/t product", "Value": 16.3},
    ("iron_dri", "Pellets_high", "Coal", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 298.4,
    },
    ("iron_dri", "Pellets_high", "Coal", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.328714157706093,
    },
    ("iron_dri", "Pellets_high", "Hydrogen", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 384.55693195398425,
    },
    ("iron_dri", "Pellets_high", "Hydrogen", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 120.0,
    },
    ("iron_dri", "Pellets_high", "Hydrogen", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "GJ/t product",
        "Value": 5.666592,
    },
    ("iron_dri", "Pellets_high", "Hydrogen", "Energy", "Natural gas", "Other"): {
        "Unit": "GJ/t product",
        "Value": 0.8821620295268815,
    },
    ("iron_dri", "Pellets_high", "Hydrogen", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.328714157706093,
    },
    ("iron_dri", "Pellets_high", "Natural gas", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 120.0,
    },
    ("iron_dri", "Pellets_high", "Natural gas", "Energy", "Natural gas", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 0.0,
    },
    ("iron_dri", "Pellets_high", "Natural gas", "Energy", "Natural gas", "Reductant"): {
        "Unit": "GJ/t product",
        "Value": 11.0,
    },
    ("iron_dri", "Pellets_high", "Natural gas", "Materials", "Pellets_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.328714157706093,
    },
    ("iron_dri", "Pellets_high", "null", "Constraint", "Pellets_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_dri", "Pellets_high", "null", "Constraint", "Pellets_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_dri", "Pellets_mid", "Coal", "Energy", "Coal", "Heat"): {"Unit": "GJ/t product", "Value": 0.0},
    ("iron_dri", "Pellets_mid", "Coal", "Energy", "Coal", "Reductant"): {"Unit": "GJ/t product", "Value": 16.3},
    ("iron_dri", "Pellets_mid", "Coal", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 298.4,
    },
    ("iron_dri", "Pellets_mid", "Coal", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.2845098566308244,
    },
    ("iron_dri", "Pellets_mid", "Hydrogen", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 347.8908907688144,
    },
    ("iron_dri", "Pellets_mid", "Hydrogen", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 120.0,
    },
    ("iron_dri", "Pellets_mid", "Hydrogen", "Energy", "Hydrogen", "Reductant"): {
        "Unit": "GJ/t product",
        "Value": 5.126304000000001,
    },
    ("iron_dri", "Pellets_mid", "Hydrogen", "Energy", "Natural gas", "Other"): {
        "Unit": "GJ/t product",
        "Value": 0.7715012824817205,
    },
    ("iron_dri", "Pellets_mid", "Hydrogen", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.2845098566308244,
    },
    ("iron_dri", "Pellets_mid", "Natural gas", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 120.0,
    },
    ("iron_dri", "Pellets_mid", "Natural gas", "Energy", "Natural gas", "Heat"): {"Unit": "GJ/t product", "Value": 0.0},
    ("iron_dri", "Pellets_mid", "Natural gas", "Energy", "Natural gas", "Reductant"): {
        "Unit": "GJ/t product",
        "Value": 11.000000000000002,
    },
    ("iron_dri", "Pellets_mid", "Natural gas", "Materials", "Pellets_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.2845098566308244,
    },
    ("iron_dri", "Pellets_mid", "null", "Constraint", "Pellets_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_dri", "Pellets_mid", "null", "Constraint", "Pellets_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_dri", "null", "Coal", "Constraint", "Coal", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_dri", "null", "Coal", "Constraint", "Coal", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_dri", "null", "Hydrogen", "Constraint", "Hydrogen", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_dri", "null", "Hydrogen", "Constraint", "Hydrogen", "Minium"): {"Unit": "%", "Value": 0.0},
    ("iron_dri", "null", "Natural gas", "Constraint", "Natural gas", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("iron_dri", "null", "Natural gas", "Constraint", "Natural gas", "Minium"): {"Unit": "%", "Value": 0.0},
    ("prep_coke", "null", "Coking coal", "Constraint", "Coking coal", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("prep_coke", "null", "Coking coal", "Constraint", "Coking coal", "Minium"): {"Unit": "%", "Value": 1.0},
    ("prep_coke", "null", "Coking coal", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t coke",
        "Value": 5.555555555555555,
    },
    ("prep_coke", "null", "Coking coal", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t coke",
        "Value": 3.9800000000000004,
    },
    ("prep_coke", "null", "Coking coal", "Materials", "Coking coal", "Feedstock"): {"Unit": "t/t coke", "Value": 1.22},
    ("prep_pellet", "IO_high", "null", "Constraint", "IO_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("prep_pellet", "IO_high", "null", "Constraint", "IO_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("prep_pellet", "IO_high", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t pellet",
        "Value": 41.0,
    },
    ("prep_pellet", "IO_high", "null", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t pellet", "Value": 0.42},
    ("prep_pellet", "IO_high", "null", "Materials", "IO_high", "Feedstock"): {"Unit": "t/t pellet", "Value": 1.0},
    ("prep_pellet", "IO_mid", "null", "Constraint", "IO_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("prep_pellet", "IO_mid", "null", "Constraint", "IO_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("prep_pellet", "IO_mid", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t pellet",
        "Value": 41.0,
    },
    ("prep_pellet", "IO_mid", "null", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t pellet", "Value": 1.06},
    ("prep_pellet", "IO_mid", "null", "Materials", "IO_mid", "Feedstock"): {"Unit": "t/t pellet", "Value": 1.0},
    ("steel_bof", "DRI_high", "null", "Constraint", "DRI_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "DRI_high", "null", "Constraint", "DRI_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "DRI_high", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 37.44360073652292,
    },
    ("steel_bof", "DRI_high", "null", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.2107134448283747,
    },
    ("steel_bof", "DRI_high", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.03675880284439286,
    },
    ("steel_bof", "DRI_high", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.10660052824873928,
    },
    ("steel_bof", "DRI_high", "null", "Materials", "DRI_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.177116256203829,
    },
    ("steel_bof", "DRI_mid", "null", "Constraint", "DRI_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "DRI_mid", "null", "Constraint", "DRI_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "DRI_mid", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 44.610291894000206,
    },
    ("steel_bof", "DRI_mid", "null", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.4424435447283803,
    },
    ("steel_bof", "DRI_mid", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.12365484407093384,
    },
    ("steel_bof", "DRI_mid", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.35859904780570817,
    },
    ("steel_bof", "DRI_mid", "null", "Materials", "DRI_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.4024158667840196,
    },
    ("steel_bof", "HBI_high", "null", "Constraint", "HBI_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "HBI_high", "null", "Constraint", "HBI_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "HBI_high", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 37.44360073652292,
    },
    ("steel_bof", "HBI_high", "null", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 1.7112671408561018,
    },
    ("steel_bof", "HBI_high", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.03675880284439286,
    },
    ("steel_bof", "HBI_high", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.10660052824873928,
    },
    ("steel_bof", "HBI_high", "null", "Materials", "HBI_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.177116256203829,
    },
    ("steel_bof", "HBI_mid", "null", "Constraint", "HBI_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "HBI_mid", "null", "Constraint", "HBI_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "HBI_mid", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 44.610291894000206,
    },
    ("steel_bof", "HBI_mid", "null", "Energy", "Flexible", "Heat"): {
        "Unit": "GJ/t product",
        "Value": 2.0388030307068954,
    },
    ("steel_bof", "HBI_mid", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.12365484407093384,
    },
    ("steel_bof", "HBI_mid", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.35859904780570817,
    },
    ("steel_bof", "HBI_mid", "null", "Materials", "HBI_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.4024158667840196,
    },
    ("steel_bof", "Hot metal", "null", "Constraint", "Hot metal", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "Hot metal", "null", "Constraint", "Hot metal", "Minium"): {"Unit": "%", "Value": 0.7},
    ("steel_bof", "Hot metal", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 34.85817316437359,
    },
    ("steel_bof", "Hot metal", "null", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 0.0},
    ("steel_bof", "Hot metal", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.008615554298366596,
    },
    ("steel_bof", "Hot metal", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.024985107465263126,
    },
    ("steel_bof", "Hot metal", "null", "Materials", "Hot metal", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.095838046722067,
    },
    ("steel_bof", "Pig iron", "null", "Constraint", "Pig iron", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "Pig iron", "null", "Constraint", "Pig iron", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "Pig iron", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 34.85817316437359,
    },
    ("steel_bof", "Pig iron", "null", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 0.0},
    ("steel_bof", "Pig iron", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.008615554298366596,
    },
    ("steel_bof", "Pig iron", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.024985107465263126,
    },
    ("steel_bof", "Pig iron", "null", "Materials", "Pig iron", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.095838046722067,
    },
    ("steel_bof", "Scrap", "null", "Constraint", "Scrap", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_bof", "Scrap", "null", "Constraint", "Scrap", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_bof", "Scrap", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 32.62745690056601,
    },
    ("steel_bof", "Scrap", "null", "Energy", "Flexible", "Heat"): {"Unit": "GJ/t product", "Value": 1.4911572013739554},
    ("steel_bof", "Scrap", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.004838525964379703,
    },
    ("steel_bof", "Scrap", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.014031725296701138,
    },
    ("steel_bof", "Scrap", "null", "Materials", "Scrap", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.0257109135015452,
    },
    ("steel_eaf", "DRI_high", "null", "Constraint", "DRI_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "DRI_high", "null", "Constraint", "DRI_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "DRI_high", "null", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 326.4081480016136,
    },
    ("steel_eaf", "DRI_high", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 37.44360073652292,
    },
    ("steel_eaf", "DRI_high", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.03675880284439286,
    },
    ("steel_eaf", "DRI_high", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.10660052824873928,
    },
    ("steel_eaf", "DRI_high", "null", "Materials", "DRI_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.177116256203829,
    },
    ("steel_eaf", "DRI_mid", "null", "Constraint", "DRI_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "DRI_mid", "null", "Constraint", "DRI_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "DRI_mid", "null", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 411.9179525314357,
    },
    ("steel_eaf", "DRI_mid", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 44.610291894000206,
    },
    ("steel_eaf", "DRI_mid", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.12365484407093384,
    },
    ("steel_eaf", "DRI_mid", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.35859904780570817,
    },
    ("steel_eaf", "DRI_mid", "null", "Materials", "DRI_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.4024158667840196,
    },
    ("steel_eaf", "HBI_high", "null", "Constraint", "HBI_high", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "HBI_high", "null", "Constraint", "HBI_high", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "HBI_high", "null", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 420.4856843930561,
    },
    ("steel_eaf", "HBI_high", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 37.44360073652292,
    },
    ("steel_eaf", "HBI_high", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.03675880284439286,
    },
    ("steel_eaf", "HBI_high", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.10660052824873928,
    },
    ("steel_eaf", "HBI_high", "null", "Materials", "HBI_high", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.177116256203829,
    },
    ("steel_eaf", "HBI_mid", "null", "Constraint", "HBI_mid", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "HBI_mid", "null", "Constraint", "HBI_mid", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "HBI_mid", "null", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 525.8011511724798,
    },
    ("steel_eaf", "HBI_mid", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 44.610291894000206,
    },
    ("steel_eaf", "HBI_mid", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.12365484407093384,
    },
    ("steel_eaf", "HBI_mid", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.35859904780570817,
    },
    ("steel_eaf", "HBI_mid", "null", "Materials", "HBI_mid", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.4024158667840196,
    },
    ("steel_eaf", "Hot metal", "null", "Constraint", "Hot metal", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "Hot metal", "null", "Constraint", "Hot metal", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "Hot metal", "null", "Energy", "Electricity", "Heat"): {"Unit": "kWh/t product", "Value": 0.0},
    ("steel_eaf", "Hot metal", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 34.85817316437359,
    },
    ("steel_eaf", "Hot metal", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.008615554298366596,
    },
    ("steel_eaf", "Hot metal", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.024985107465263126,
    },
    ("steel_eaf", "Hot metal", "null", "Materials", "Hot metal", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.095838046722067,
    },
    ("steel_eaf", "Pig iron", "null", "Constraint", "Pig iron", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "Pig iron", "null", "Constraint", "Pig iron", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "Pig iron", "null", "Energy", "Electricity", "Heat"): {"Unit": "kWh/t product", "Value": 0.0},
    ("steel_eaf", "Pig iron", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 34.85817316437359,
    },
    ("steel_eaf", "Pig iron", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.008615554298366596,
    },
    ("steel_eaf", "Pig iron", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.024985107465263126,
    },
    ("steel_eaf", "Pig iron", "null", "Materials", "Pig iron", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.095838046722067,
    },
    ("steel_eaf", "Scrap", "null", "Constraint", "Scrap", "Maximum"): {"Unit": "%", "Value": 1.0},
    ("steel_eaf", "Scrap", "null", "Constraint", "Scrap", "Minium"): {"Unit": "%", "Value": 0.0},
    ("steel_eaf", "Scrap", "null", "Energy", "Electricity", "Heat"): {
        "Unit": "kWh/t product",
        "Value": 339.68561047298704,
    },
    ("steel_eaf", "Scrap", "null", "Energy", "Electricity", "Machine drive"): {
        "Unit": "kWh/t product",
        "Value": 32.62745690056601,
    },
    ("steel_eaf", "Scrap", "null", "Materials", "Burnt dolomite", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.004838525964379703,
    },
    ("steel_eaf", "Scrap", "null", "Materials", "Burnt lime", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 0.014031725296701138,
    },
    ("steel_eaf", "Scrap", "null", "Materials", "Scrap", "Feedstock"): {
        "Unit": "t/t product",
        "Value": 1.0257109135015452,
    },
}



from typing import Callable


class EnvironmentScorer:
    @staticmethod
    def normalize_and_score(
        grid: list[dict[str, float]],
        distance_columns: list[str],
        scaler: Callable[[list[list[float]]], list[list[float]]],
    ) -> list[dict[str, float]]:
        """
        Normalize distances and compute proximity scores.

        Args:
            grid (list[dict[str, float]]): The grid with distance columns as a list of dictionaries.
            distance_columns (list[str]): list of column names to normalize.
            scaler (Callable): A scaling function provided by adapters.

        Returns:
            list[dict[str, float]]: Grid with normalized and scored data as a list of dictionaries.
        """
        # Prepare data for scaling
        data = [[row[col] for col in distance_columns] for row in grid]
        normalized = scaler(data)

        for i, row in enumerate(grid):
            for j, col in enumerate(distance_columns):
                row[f"{col}_norm"] = normalized[i][j]

            row["average_score"] = sum(row[f"{col}_norm"] for col in distance_columns) / len(distance_columns)
            row["proximity_score"] = 1 - row["average_score"]

        return grid



from typing import Dict

# TECHNOLOGY GROUPINGS

MYPY_DICT_STR_LIST = Dict[str, list]

FURNACE_GROUP_DICT: MYPY_DICT_STR_LIST = {
    "blast_furnace": [
        "Avg BF-BOF",
        "BAT BF-BOF",
        "BAT BF-BOF_bio PCI",
        "BAT BF-BOF_H2 PCI",
        "BAT BF-BOF+CCUS",
        "BAT BF-BOF+BECCUS",
        "BAT BF-BOF+CCU",
    ],
    "dri-bof": ["DRI-Melt-BOF", "DRI-Melt-BOF_100% zero-C H2", "DRI-Melt-BOF+CCUS"],
    "dri-eaf": [
        "DRI-EAF",
        "DRI-EAF_50% bio-CH4",
        "DRI-EAF_50% green H2",
        "DRI-EAF+CCUS",
        "DRI-EAF_100% green H2",
    ],
    "smelting_reduction": ["Smelting Reduction", "Smelting Reduction+CCUS"],
    "eaf-basic": ["EAF"],
    "eaf-advanced": ["Electrolyzer-EAF", "Electrowinning-EAF"],
    "ccs": [
        "BAT BF-BOF+BECCUS",
        "BAT BF-BOF+CCUS",
        "DRI-Melt-BOF+CCUS",
        "DRI-EAF+CCUS",
        "Smelting Reduction+CCUS",
    ],
    "ccu": ["BAT BF-BOF+CCU"],
}
FURNACE_GROUP_DICT["dri"] = FURNACE_GROUP_DICT["dri-bof"] + FURNACE_GROUP_DICT["dri-eaf"]
FURNACE_GROUP_DICT["eaf-all"] = FURNACE_GROUP_DICT["eaf-basic"] + FURNACE_GROUP_DICT["eaf-advanced"]



import logging

from ..service_layer.message_bus import MessageBus
from ..domain.trade_modelling.hitchcock_modelling import steel_trade_HP
from ..domain.events import SteelAllocationsCalculated

logger = logging.getLogger(__name__)


class AllocationModel:
    """Model to allocate resource to demand center from supply center"""

    @staticmethod
    def run(bus: MessageBus) -> None:
        """
        Run the trade model

        Args

        """
        allocations = steel_trade_HP(bus.uow.repository, bus.env.year)
        if (event := SteelAllocationsCalculated(allocations=allocations)) is not None:
            bus.handle(event)


class PlantAgentsModel:
    """Simplest possible economic model for a plant agent."""

    @staticmethod
    def run(bus: MessageBus) -> None:
        """
        Run the plant agent model.

        Args:
            bus (MessageBus): The message bus to send the event to
        """

        plants = bus.uow.plants.list()
        for plant in plants:
            for fg in plant.furnace_groups:
                if (
                    cmd := plant.evaluate_furnace_group_strategy(
                        fg.furnace_group_id, price=bus.env.extract_price_from_costcurve(demand=bus.env.current_demand)
                    )
                ) is not None:
                    bus.handle(cmd)

            # Maybe not the prettiest but it works?
            try:  # TODO: @Marcus - This will be updated as part of ticket on new expansion strategy
                assert isinstance(plant.location.iso3, str)
                new_furnace = plant.generate_new_furnace(
                    technology_name=bus.env.new_furnace_npv()[1],
                    current_year=bus.env.year,
                    product="steel",
                    # FIXME: workaround iso3 key error
                    cost_of_debt=bus.env.cost_of_capital.get(plant.location.iso3, 0.05),
                )  # TODO: cost of debt should come from plant level since location specific
                if (cmd := plant.evaluate_expansion(bus.env, new_furnace=new_furnace)) is not None:
                    bus.handle(cmd)
            except ValueError as e:
                if "Cost curve is empty" in str(e):
                    logger.warning("Cannot evaluate expansion: Cost curve is empty")
                else:
                    # Re-raise other ValueErrors
                    raise



"""
This package contains mapping data for various domain concepts.

It provides dictionaries and utility functions for accessing reference data
that is used throughout the application.
"""

from .country import (
    COUNTRY_TO_CODE_MAP,
    COUNTRY_TO_IRENA_MAP,
    COUNTRY_TO_REGION_MAP,
    COUNTRY_TO_SSP_REGION_MAP,
    GEM_COUNTRY_WS_REGION_MAP,
)

__all__ = [
    "COUNTRY_TO_CODE_MAP",
    "COUNTRY_TO_IRENA_MAP",
    "COUNTRY_TO_REGION_MAP",
    "COUNTRY_TO_SSP_REGION_MAP",
    "GEM_COUNTRY_WS_REGION_MAP",
]
